#include <stdio.h>
#include <zlib.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>

#define MAQ_LONGREADS

#ifdef MAQ_LONGREADS
#  define MAX_READLEN 128
#else
#  define MAX_READLEN 64
#endif

#define MAX_NAMELEN 36
#define MAQMAP_FORMAT_OLD 0
#define MAQMAP_FORMAT_NEW -1

#define PAIRFLAG_FF      0x01
#define PAIRFLAG_FR      0x02
#define PAIRFLAG_RF      0x04
#define PAIRFLAG_RR      0x08
#define PAIRFLAG_PAIRED  0x10
#define PAIRFLAG_DIFFCHR 0x20
#define PAIRFLAG_NOMATCH 0x40
#define PAIRFLAG_SW      0x80

/*
  "130" shows the alternative definition when flag==(PAIRFLAG_SW|PAIRFLAG_PAIRED).

  name: read name
  size: the length of the read
  seq: read sequence + qualities (see also below)
  seq[MAX_READLEN-1]: SE mapQ (equals map_qual if unpaired); 130: indel length
  map_qual: the final mapping quality; 130: position of the indel (0 if no indel)
  alt_qual: the lower mapQ of the two ends (equals map_qual if unpaired); 130: mapQ of its mate
  flag: status of the pair, a bitwise flag using PAIRFLAG_ macros
  dist: offset of the mate (zero if unpaired, or two ends mapped to different chr)
  info1: mismatches in the 28bp (higher 4 bits) and mismatches (lower 4 bits)
  info2: sum of errors of the best hit
  c[2]: counts of all 0- and 1-mismatch hits on the reference
 */
typedef struct {
	uint8_t seq[MAX_READLEN]; /* the last base is the single-end mapping quality. */
	uint8_t size, map_qual, info1, info2, c[2], flag, alt_qual;
	uint32_t seqid, pos;
	int dist;
	char name[MAX_NAMELEN];
} maqmap1_t;

typedef struct {
	int format, n_ref;
	char **ref_name;
} maqmap_t;

maqmap_t *maqmap_read_header(gzFile fp)
{
	maqmap_t *mm;
	int k, len;
	uint64_t tmp = 0;
	mm = (maqmap_t*)calloc(1, sizeof(maqmap_t));
	gzread(fp, &mm->format, sizeof(int));
	if (mm->format != MAQMAP_FORMAT_NEW) {
		fprintf(stderr, "Wrong alignment format!\n");
		exit(3);
	}
	gzread(fp, &mm->n_ref, sizeof(int));
	mm->ref_name = (char**)calloc(mm->n_ref, sizeof(char*));
	for (k = 0; k != mm->n_ref; ++k) {
		gzread(fp, &len, sizeof(int));
		mm->ref_name[k] = (char*)malloc(len * sizeof(char));
		gzread(fp, mm->ref_name[k], len);
	}
	gzread(fp, &tmp, 8); /* for backward compatibility */
	return mm;
}

void maqmap_write_header(gzFile fp, const maqmap_t *mm)
{
	int i, len;
	uint64_t tmp;
	gzwrite(fp, &mm->format, sizeof(int));
	gzwrite(fp, &mm->n_ref, sizeof(int));
	for (i = 0; i != mm->n_ref; ++i) {
		len = strlen(mm->ref_name[i]) + 1;
		gzwrite(fp, &len, sizeof(int));
		gzwrite(fp, mm->ref_name[i], len);
	}
	gzwrite(fp, &tmp, 8); /* useless, for backward compatibility */
}

void maq_delete_maqmap(maqmap_t *mm)
{
	int i;
	if (mm == 0) return;
	for (i = 0; i < mm->n_ref; ++i) free(mm->ref_name[i]);
	free(mm->ref_name);
	free(mm);
}

void mapview_core(gzFile fpin)
{
	uint32_t j;
	maqmap_t *m = maqmap_read_header(fpin);
	maqmap1_t *m1, mm1;
	m1 = &mm1;
	while (gzread(fpin, m1, sizeof(maqmap1_t)) == sizeof(maqmap1_t)) {
		printf("%s\t%s\t%d\t%c\t%d\t%u\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d",
			   m1->name, m->ref_name[m1->seqid], (m1->pos>>1) + 1,
			   (m1->pos&1)? '-' : '+', m1->dist, m1->flag, m1->map_qual,
			   (signed char)m1->seq[MAX_READLEN-1], m1->alt_qual,
			   m1->info1&0xf, m1->info2, m1->c[0], m1->c[1], m1->size);
		putchar('\t');
		for (j = 0; j != m1->size; ++j) {
			if (m1->seq[j] == 0) putchar('n');
			else putchar("ACGT"[m1->seq[j]>>6&3]);
		}
		putchar('\t');
		for (j = 0; j != m1->size; ++j)
			putchar((m1->seq[j]&0x3f) + 33);
		putchar('\n');
	}
	maq_delete_maqmap(m);
}

int main(int argc, char *argv[])
{
	gzFile fp;
	if (argc == 1) {
		fprintf(stderr, "Usage: mapview <in.map>\n");
		return 1;
	}
	fp = (strcmp(argv[1], "-") == 0)? gzdopen(fileno(stdin), "r") : gzopen(argv[1], "r");
	mapview_core(fp);
	gzclose(fp);
	return 0;
}
